import numpy as np
import yaml
import pandas as pd
from disc_fit import analysis,calling_all
import os 
import matplotlib.pyplot as plt

def read_data(input_path):
    """
    Reads in data from a text file. USER DEFINED.

    Parameters
    ----------
    input_path : str
        Path to the input file.

    Returns
    -------
    xdata : np.ndarray
        Array of x-data.
    ydata : np.ndarray
        Array of y-data.
    yerr : np.ndarray
        Array of y-errors.

    """
    xdata,ydata,yerr = np.loadtxt(input_path,unpack=True,usecols=[0,1,2])

    indsort = np.argsort(xdata)
    xdata = xdata[indsort]
    ydata = ydata[indsort]
    yerr = yerr[indsort]

    xdata = 1e15*xdata 
    ydata = 1e45*ydata
    yerr = 1e45*yerr

    return xdata,ydata,yerr


def read_yaml(file_path):
    ''' 
    Function
    ----------------------------------------------------------------------
    ----------------------------------------------------------------------
    read yaml (i.e. configuration) file
        
        Parameters:
            file_path: str
                absolute path to read
        Returns:
    '''
    with open(file_path, "r") as f:
        return yaml.safe_load(f)
    

def read_config_and_launch(config_path,read_data):
    """
    This function reads in a YAML configuration file and launches the 
    parameter estimation process.

    Parameters
    ----------
    config_path : str
        Path to the YAML configuration file.
    read_data: function
        A user defined function that takes in a file path and returns xdata, ydata, and yerr.

        
    Returns
    -------
    post_df : pandas.DataFrame
        Posterior samples from the MCMC sampler.
    best : dict
        Best-fit parameters and their uncertainties.

    """

    config_path = './config.yaml'

    config = read_yaml(config_path)

    if config['Hyperparameter_estimator']['raynest']['use'] == True:
        hyperpar_dict = config['Hyperparameter_estimator']['raynest']
        which_sampler ='raynest'

    elif config['Hyperparameter_estimator']['CPnest']['use'] == True:
        hyperpar_dict = config['Hyperparameter_estimator']['CPnest']
        which_sampler ='CPnest'
    elif config['Hyperparameter_estimator']['nessai']['use'] == True:
        hyperpar_dict = config['Hyperparameter_estimator']['nessai']
        which_sampler ='nessai'
    else:
        print('Not using Nested Sampling, no other estimator is available, exiting')
        exit()

    fitting_parameters = config['Parameters']
    hyperparameter_disc_model = config['Hyperparameter_disc_model']

    names = []
    LaTex_names = []
    bounds = []
    vary = []
    value = []

    for k in fitting_parameters:
        names.append(fitting_parameters[k]['name'])
        LaTex_names.append(fitting_parameters[k]["LaTex_name"])
        bounds.append(fitting_parameters[k]["bounds"])
        vary.append(fitting_parameters[k]["vary"])
        value.append(fitting_parameters[k]["value"])  


    variable_quantities = {names[i]:bounds[i] for i in range(0,len(names)) if vary[i]==True}
    fixed_quantities = {names[i]:value[i] for i in range(0,len(names)) if vary[i]==False}
    variable_LaTex = {names[i]:LaTex_names[i] for i in range(0,len(names)) if vary[i]==True}
    fixed_LaTex = {names[i]:LaTex_names[i] for i in range(0,len(names)) if vary[i]==False}

    names_hyp = []
    value_hyp = []

    for k in hyperparameter_disc_model:
        names_hyp.append(hyperparameter_disc_model[k]['name'])
        value_hyp.append(hyperparameter_disc_model[k]["value"]) 

    hyperpar_disc_dict = {names_hyp[i]:value_hyp[i] for i in range(0,len(names_hyp))}

    input_path = config['Settings']['input_path']
    output_path = config['Settings']['output_path']

    xdata,ydata,yerr = read_data(input_path)

    if os.path.exists(output_path+"posterior.csv"):
        print('Parameter estimation file already exists, plotting diagnostics')

        post_df = pd.read_csv(output_path+"posterior.csv")
        best = analysis(post_df,xdata,ydata,yerr,output_path,fixed_quantities,\
                        variable_LaTex,hyperpar_disc_dict)

    else:    
        post_df,best = calling_all(xdata,ydata,yerr,output_path,hyperpar_dict,\
                    variable_quantities,fixed_quantities,variable_LaTex,\
                    hyperpar_disc_dict,which_sampler)
        
    return post_df,best


if __name__=='__main__':
    config_path = './config.yaml'

    post_df,best = read_config_and_launch(config_path,read_data)

    plt.show()


import numpy as np
import matplotlib.pyplot as plt
from scipy.integrate import quad,quad_vec,dblquad


G =  6.67408e-8                 # gravitational constant in cm^3 g^-1 s^-2
m_p = 1.6726219e-24             # proton mass in g
c = 2.99792458e10               # light speed of light in cm s^-1
L_sun = 3.828e33                # solar luminosity in erg s^-1
R_sun = 6.95700000e10           # solar radius in cm
sigma = 5.670367e-5             # stefan-boltzmann constant in erg cm^−2 s^−1 K^−4
h = 6.62607004e-27              # planck constant in erg s
k = 1.38064852e-16              # Boltzman const erg K−1
M_sun = 1.98847542e33           # mass of sun in g
sigma_tmp = 6.65245872e-25      # thompson cross section in cm^2


def r_swarz(m):
    """
    Calculates the Schwarzschild radius of a given mass.

    Parameters
    ----------
    m : float
        The mass of the object.

    Returns
    -------
    r_schw : float
        The Schwarzschild radius of the object.

    """
    return (2 * G * m) / (c ** 2)

def l_edd (m):
    """
    Calculates the luminosity of an Eddington-Rutherford disk at a given mass.

    Parameters
    ----------
    m : float
        The mass of the disk in solar masses.

    Returns
    -------
    lum : float
        The luminosity of the disk in erg/s.

    """
    return ( ( 4 * np.pi * c * G * m * m_p )/ sigma_tmp )




def temperatura (r, m,m_dot,r_in):
    """
    Calculates the temperature of alpha-disc.

    Parameters
    ----------
    r : float / numpy array
        The distance from the center.
    m : float
        The mass of the object.
    m_dot : float
        Accretion rate.
    r_in : float
        The inner radius of the potential well.

    Returns
    -------
    T : float / numpy array
        The temperature of the object.

    """
    a = 3 * G * m * m_dot
    b = 8 * np.pi * sigma * r**3
    d = (1 - (r_in/r)**0.5)
    
    return ((a/b)*(d))**0.25 


def brightness(nu,T):
    """
    Calculates the blackbody radiation intensity at a given frequency and temperature.

    Parameters
    ----------
    nu: float / numpy array
        Frequency in Hz.
    T: float 
        Temperature in Kelvin, see temperatura(r,m,m_dot,r_in)

    Returns
    -------
    br: float
        Blackbody radiation intensity at the given frequency and temperature.

    """
    return ((2*h/(c**2))*(nu**3/(np.exp(h*nu/(k*T))-1)))


def integrand (r, nu, m, m_dot, r_in) :
    """
    Calculates the integral of the radiation intensity over the radius.

    Parameters
    ----------
    r : float / numpy array
        Distance from the center.
    nu : float
        Frequency in Hz.
    m : float
        The mass of the object.
    m_dot : float
        Accretion rate.
    r_in : float
        The inner radius of the potential well.

    Returns
    -------
    I : float
        The integral of the radiation intensity over the radius.

    """
    return  4*np.pi*np.pi*r*brightness(nu,temperatura(r,m,m_dot,r_in))


def disc_luminosity(freq,m,m_dot,theta, r_in=3,r_fin=1e4) : 
    """
    Calculates the luminosity of an accretion disk at a given frequency, mass, accretion rate, and viewing angle.

    Parameters
    ----------
    freq : float or numpy.ndarray
        Frequency of the radiation in Hz.
    m : float
        The mass of the disk in log10 solar masses.
    m_dot : float
        The accretion rate of the disk in log10 Eddington luminosity.
    theta : float
        The angle between the line of sight and the major axis of the disk.
    r_in : float, optional
        Inner radius of the accretion disk in unit of Schwarzschild radii. The default is 3. 
    r_fin : float, optional
        Outer radius of the accretion disk in unit of Schwarzschild radii. The default is 1e4.
    Returns
    -------
    lum : float or numpy.ndarray
        The log10 luminosity of the disk multiplied by the frequency.

    """
    m = 10**m * M_sun
   
    m_dot =  10**m_dot*l_edd(m)/(c**2)

    rs = r_swarz(m)

    r_in *= rs
    r_fin *= rs 

    integrand2 = lambda r : integrand(r,freq,m,m_dot,r_in)

    lumx,_ = quad_vec(integrand2,r_in,r_fin)
    
    return np.log10(freq*lumx*2*np.cos(theta))


def disc_bol_luminosity(m,m_dot, r_in=3,r_fin=1e4,nu_min=10**10,nu_max=10**16.5): 
    """
    Calculates the bolometric luminosity of an accretion disk at a given mass and accretion rate.

    Parameters
    ----------
    m : float
        The mass of the disk in log10 solar masses.
    m_dot : float
        The accretion rate of the disk in log10 Eddington luminosities.
    r_in : float, optional
        Inner radius of the accretion disk in unit of Schwarzschild radii. The default is 3. 
    r_fin : float, optional
        Outer radius of the accretion disk in unit of Schwarzschild radii. The default is 1e4.
    nu_min : float, optional
        Minimum frequency for integrating the bolometric luminosity. The default is 10**10.
    nu_max : float, optional
        Maximum frequency for integrating the bolometric luminosity. The default is 10**16.5.

    Returns
    -------
    lum : float
        The bolometric luminosity of the disk in erg/s integrated over radius and frequency.
    bol_lum : float
        The bolometric luminosity of the disk in erg/s computed as eta * mdot* c^2 [eta 0.083, i.e. Rin = 3Rs]
    eta : float
        The efficiency of the disk, defined as the ratio of the bolometric luminosity to the Eddington luminosity.

    """
    m = 10**m * M_sun

    m_dot =  10**m_dot*l_edd(m)/(c**2)


    rs = r_swarz(m)
    r_in *= rs
    r_fin *= rs

    integrand2 = lambda r,freq : integrand(r,freq,m,m_dot,r_in)

    integrand3 = lambda freq : quad(integrand2,r_in,r_fin,args=(freq))[0]

    lumbol = quad(integrand3,nu_min,nu_max)

    lumx = 0.083*m_dot*c*c
    eta = lumbol[0]/m_dot/c/c

    return lumbol[0],lumx,eta




if __name__=='__main__':


    freq = 10**np.linspace(12,16,150)
    
    m = np.log10(1e9)
    m_dot =  np.log10(0.1)

    lum1  = disc_luminosity(freq,8,np.log10(1),0)
    lum2  = disc_luminosity(freq,9,np.log10(0.1),0)
    lum3  = disc_luminosity(freq,10,np.log10(0.01),0)
    lum5  = disc_luminosity(freq,11,np.log10(0.001),0)
    lum4  = disc_luminosity(freq,np.log10(3e9),np.log10(0.1/3),0)

    y = np.linspace(40,45,150)

    fig = plt.figure()
    ax  = fig.add_subplot()
    ax.plot(np.log10(freq),lum1,'r')
    ax.plot(np.log10(freq),lum2,'b')
    ax.plot(np.log10(freq),lum3,'k')
    ax.plot(np.log10(freq),lum4,'m')
    ax.plot(np.log10(freq),lum5,'c')
    #ax.plot(np.log10(5.88e14)*np.ones_like(y),y,'c')

    plt.ylim(40,46)
    plt.show()


    lum  = disc_luminosity(freq,m,m_dot,0)
    lum_bol,bol_lum,eta  = disc_bol_luminosity(m,m_dot)

    


import numpy as np
import raynest.model
from disc_model import disc_luminosity #,disco_bol_luminosity,disco_peak_nuLnu
import matplotlib.pyplot as plt
import yaml
import os
import pandas as pd
import cpnest.model
import nessai.model
from nessai.flowsampler import FlowSampler
import h5py

def likelihood(ydata,ymodel,yerr):
    """
    Calculates the likelihood function for a set of data and a model.

    Parameters
    ----------
    ydata : array_like
        The observed data values.
    ymodel : array_like
        The model predictions.
    yerr : array_like
        The uncertainty in the observed data.

    Returns
    -------
    lk : float
        The calculated likelihood value.

    """
    chi2 = (ydata-ymodel)/yerr
    lk   = chi2*chi2 + np.log(2*np.pi*yerr*yerr)

    return -0.5*np.sum(lk)

class Quasar_rayCP(raynest.model.Model):
    """
    Quasar class for accretion disc model raynest or CPnest.

    Parameters
    ----------
    freq : array_like
        Frequency array.
    lum : array_like
        Observed luminosity array. In solar luminosities.
    lum_err : array_like
        Uncertainty in the observed luminosity array. In solar luminosities.

    Attributes
    ----------
    freq : array_like
        Frequency array.
    lum : array_like
        Observed luminosity array. In solar luminosities.
    lum_err : array_like
        Uncertainty in the observed luminosity array. In solar luminosities.

    """

    names = []
    bounds = []

    def __init__(self,
                 freq,
                 lum,
                 lum_err,
                 variable_quantities,
                 fixed_quantities,
                 hyperpar_disc_dict,
                 *args,**kwargs):

        super(Quasar_rayCP,self).__init__()

        self.fixed_quantities    = fixed_quantities
        self.variable_quantities = variable_quantities
        self.bounds      = np.array(list(self.variable_quantities.values()))
        self.names       = list(self.variable_quantities.keys())
        self.names_fixed = list(fixed_quantities.keys())
        self.value       = list(fixed_quantities.values())     


        self.freq = freq 
        self.lum = np.log10(lum) 
        self.lum_err = lum_err/(lum*np.log(10)) 

        self.Rin = hyperpar_disc_dict['Rin']
        self.Rfin = hyperpar_disc_dict['Rfin']
        self.nu_min = hyperpar_disc_dict['nu_min']
        self.nu_max = hyperpar_disc_dict['nu_max']
   

    def log_likelihood(self,p):
        """
        Calculates the log-likelihood function for a set of data and a model.

        Parameters
        ----------
        p : dict
            The parameter values.

        Returns
        -------
        lk : float
            The calculated log-likelihood value.

        """
        if 'log10Mbh' in self.variable_quantities:
            m = p['log10Mbh']
        else:
            m = self.fixed_quantities['log10Mbh']

        if 'log10Mdot' in self.variable_quantities:
            mdot = p['log10Mdot']
        else:
            mdot = self.fixed_quantities['log10Mdot']

        if 'theta' in self.variable_quantities:
            theta = p['theta']
        else:
            theta = self.fixed_quantities['theta']

        theta = theta*np.pi/180

        ymodel = disc_luminosity(self.freq,m,mdot,theta,r_in=self.Rin,r_fin=self.Rfin)
        lk = likelihood(self.lum,ymodel,self.lum_err)
        return  lk


class Quasar_nessai(nessai.model.Model):
    """
    Quasar class for accretion disc model nessai.

    Parameters
    ----------
    freq : array_like
        Frequency array.
    lum : array_like
        Observed luminosity array. In solar luminosities.
    lum_err : array_like
        Uncertainty in the observed luminosity array. In solar luminosities.

    Attributes
    ----------
    freq : array_like
        Frequency array.
    lum : array_like
        Observed luminosity array. In solar luminosities.
    lum_err : array_like
        Uncertainty in the observed luminosity array. In solar luminosities.

    """

    names = []
    bounds = []

    def __init__(self,
                 freq,
                 lum,
                 lum_err,
                 variable_quantities,
                 fixed_quantities,
                 hyperpar_disc_dict,
                 *args,**kwargs):

        super(Quasar_nessai,self).__init__()

        self.fixed_quantities    = fixed_quantities
        self.variable_quantities = variable_quantities
        self.bounds = self.variable_quantities
        self.names       = list(self.variable_quantities.keys())
        self.names_fixed = list(fixed_quantities.keys())
        self.value       = list(fixed_quantities.values())     

        

        self.freq = freq 
        self.lum = np.log10(lum) 
        self.lum_err = lum_err/(lum*np.log(10)) 

        self.Rin = hyperpar_disc_dict['Rin']
        self.Rfin = hyperpar_disc_dict['Rfin']
        self.nu_min = hyperpar_disc_dict['nu_min']
        self.nu_max = hyperpar_disc_dict['nu_max']
   
    def log_prior(self, p):
        """
        Calculates the log-prior probability for a set of parameters.

        Parameters
        ----------
        p : dict
            The parameter values.

        Returns
        -------
        lprior : float
            The calculated log-prior value.

        """
        if not self.in_bounds(p).any():
            return -np.inf
    
        lp = np.log(self.in_bounds(p))
        
        return lp

    def log_likelihood(self,p):
        """
        Calculates the log-likelihood function for a set of data and a model.

        Parameters
        ----------
        p : dict
            The parameter values.

        Returns
        -------
        lk : float
            The calculated log-likelihood value.

        """
        if 'log10Mbh' in self.variable_quantities:
            m = p['log10Mbh']
        else:
            m = self.fixed_quantities['log10Mbh']

        if 'log10Mdot' in self.variable_quantities:
            mdot = p['log10Mdot']
        else:
            mdot = self.fixed_quantities['log10Mdot']

        if 'theta' in self.variable_quantities:
            theta = p['theta']
        else:
            theta = self.fixed_quantities['theta']

        theta = theta*np.pi/180

        ymodel = disc_luminosity(self.freq,m,mdot,theta,r_in=self.Rin,r_fin=self.Rfin)
        lk = likelihood(self.lum,ymodel,self.lum_err)
        return  lk

def run_nessai(freq,lum,lum_err,outdir,hyperpar_dict,variable_quantities,\
            fixed_quantities,hyperpar_disc_dict):
    """
    Runs RayNest to estimate the parameters of an accretion disc model.

    Parameters
    ----------
    freq : array_like
        Frequency array.
    lum : array_like
        Observed luminosity array [nu * Lnu].
    lum_err : array_like
        Uncertainty in the observed luminosity array.
    outdir : str
        Output directory.
    hyperpar_dict : dict
        RayNest Hyperparameter dictionary.
    variable_quantities : dict
        Dictionary of variable quantities.
    fixed_quantities : dict
        Dictionary of fixed quantities.
    hyperpar_disc_dict : dict
        Hyperparameter dictionary for the disc model.

    Returns
    -------
    df : pandas.DataFrame
        Posterior samples as a pandas DataFrame.

    """
    model = Quasar_nessai(freq,lum,lum_err,variable_quantities,fixed_quantities,hyperpar_disc_dict)
    names = list(model.names)
    names.append('logL')
    names.append('logP')


    work=FlowSampler(model,
                    output=outdir,
                    nlive= hyperpar_dict['nlive'],
                    prior_sampling= hyperpar_dict['prior_sampling'],
                    n_pool= hyperpar_dict['n_pool'],
                    resume_file= hyperpar_dict['resume_file'],
                    checkpointing= hyperpar_dict['checkpointing'],
                    checkpoint_on_training= hyperpar_dict['checkpoint_on_training'],
                    checkpoint_interval= hyperpar_dict['checkpoint_interval'])
    work.run()

    posterior = work.posterior_samples

    df = pd.DataFrame(posterior,columns=list(posterior.dtype.names))

    df.to_csv(outdir+'/posterior.csv',index=False)

    return df

def run_raynest(freq,lum,lum_err,outdir,hyperpar_dict,variable_quantities,\
        fixed_quantities,hyperpar_disc_dict):
    """
    Runs RayNest to estimate the parameters of an accretion disc model.

    Parameters
    ----------
    freq : array_like
        Frequency array.
    lum : array_like
        Observed luminosity array [nu * Lnu].
    lum_err : array_like
        Uncertainty in the observed luminosity array.
    outdir : str
        Output directory.
    hyperpar_dict : dict
        RayNest Hyperparameter dictionary.
    variable_quantities : dict
        Dictionary of variable quantities.
    fixed_quantities : dict
        Dictionary of fixed quantities.
    hyperpar_disc_dict : dict
        Hyperparameter dictionary for the disc model.

    Returns
    -------
    df : pandas.DataFrame
        Posterior samples as a pandas DataFrame.

    """
    model = Quasar_rayCP(freq,lum,lum_err,variable_quantities,fixed_quantities,hyperpar_disc_dict)
    names = list(model.names)
    names.append('logL')
    names.append('logP')

    work=raynest.raynest(model,
                         verbose=hyperpar_dict['verbose'],
                         nnest=hyperpar_dict['nnest'],
                         nensemble=hyperpar_dict['nensemble'],
                         nlive=hyperpar_dict['nlive'],
                         maxmcmc=hyperpar_dict['maxmcmc'],
                         nslice=hyperpar_dict['nslice'],
                         nhamiltonian=hyperpar_dict['nhamiltonian'],
                         resume=hyperpar_dict['resume'],
                         output=outdir,
                         periodic_checkpoint_interval=hyperpar_dict['periodic_checkpoint_interval'])
    work.run(corner = False)

    posterior = work.posterior_samples

    df = pd.DataFrame(posterior,columns=list(posterior.dtype.names))

    df.to_csv(outdir+'/posterior.csv',index=False)

    return df


def run_CPnest(freq,lum,lum_err,outdir,hyperpar_dict,variable_quantities,\
               fixed_quantities,hyperpar_disc_dict):
    """
    Runs CPNest to estimate the parameters of an accretion disc model.

    Parameters
    ----------
    freq : array_like
        Frequency array.
    lum : array_like
        Observed luminosity array [nu * Lnu].
    lum_err : array_like
        Uncertainty in the observed luminosity array.
    outdir : str
        Output directory.
    hyperpar_dict : dict
        CPnest Hyperparameter dictionary.
    variable_quantities : dict
        Dictionary of variable quantities.
    fixed_quantities : dict
        Dictionary of fixed quantities.
    hyperpar_disc_dict : dict
        Hyperparameter dictionary for the disc model.

    Returns
    -------
    df : pandas.DataFrame
        Posterior samples as a pandas DataFrame.

    """

    model = Quasar_rayCP(freq,lum,lum_err,variable_quantities,fixed_quantities,hyperpar_disc_dict)
    names = list(model.names)
    names.append('logL')
    names.append('logP')

    work=cpnest.CPNest(model,
                       verbose=hyperpar_dict['verbose'],
                       nlive=hyperpar_dict['nlive'],
                       maxmcmc=hyperpar_dict['maxmcmc'],
                       nslice=hyperpar_dict['nslice'],
                       nhamiltonian=hyperpar_dict['nhamiltonian'],
                       resume=hyperpar_dict['resume'],
                       output=outdir,
                       nthreads=hyperpar_dict['nthreads'],
                       periodic_checkpoint_interval=hyperpar_dict['periodic_checkpoint_interval'])
    work.run()

    posterior = work.get_posterior_samples()

    df = pd.DataFrame(posterior,columns=list(posterior.dtype.names))

    df.to_csv(outdir+'/posterior.csv',index=False)
    
    return df


def evidence(file):
    """
    Reads the evidence from an output file generated by Nessai.

    Parameters
    ----------
    file : str
        Path to the output file.

    Returns
    -------
    evi : float
        The evidence value.

    """
    data = h5py.File(file,mode='r')
    evi = data['log_evidence'][()]

    return evi

def analysis(post_df,xdata,ydata,yerr,outdir,fixed_quantities,variable_LaTex,hyperpar_disc_dict):
    """
    Function to analyze the posterior samples and plot the results.

    Parameters
    ----------
    post_df : pandas.DataFrame
        Posterior samples as a pandas DataFrame.
    xdata : np.ndarray
        Array of x-data. Frequency.
    ydata : np.ndarray
        Array of y-data. nu*Lnu
    yerr : np.ndarray
        Array of y-errors.
    outdir : str
        Output directory.
    fixed_quantities : dict
        Dictionary of fixed quantities.
    variable_LaTex : dict
        Dictionary of variable quantities with LaTeX labels.
    hyperpar_disc_dict : dict
        Hyperparameter dictionary for the disc model.

    Returns
    -------
    best : list
        Best-fit values for the parameters.

    """

    if os.path.exists(outdir+'/result.hdf5'):
        evi = evidence(outdir+'/result.hdf5')
        np.savetxt(outdir+'/evidence.txt',np.array([evi]),header='log_evidence')
    else:
        pass

    eta = 1/4./hyperpar_disc_dict['Rin']

    xplot = 10**np.linspace(14.5,16.,300)

    if 'log10Mbh' in fixed_quantities:
        Mbh = fixed_quantities['log10Mbh']
    else:
        Mbh = post_df['log10Mbh'].values
        Nsample = Mbh.shape[0]
    if 'log10Mdot' in fixed_quantities:
        Mdot = fixed_quantities['log10Mdot']
    else:
        Mdot = post_df['log10Mdot'].values
        Nsample = Mdot.shape[0]
    if 'theta' in fixed_quantities:
        theta = fixed_quantities['theta']
    else:
        theta = post_df['theta'].values
        Nsample = theta.shape[0]

    Mbh50,Mbh16,Mbh84 = np.percentile(Mbh,[50,16,84])
    Mdot50,Mdot16,Mdot84 = np.percentile(Mdot,[50,16,84])
    lambda_edd = np.log10(10**Mdot*eta)
    lambda_edd50,lambda_edd16,lambda_edd84 = np.percentile(lambda_edd,[50,16,84])
    theta50,theta16,theta84 = np.percentile(theta,[50,16,84])

    print('log10 Mbh: {:.4f} + {:.4f} -{:.4f}'.format(Mbh50,Mbh84-Mbh50,Mbh50-Mbh16))
    print('log10 Mdot/(Ledd/c^2): {:.4f} + {:.4f} -{:.4f}'.format(Mdot50,Mdot84-Mdot50,Mdot50-Mdot16))
    print('log10 lambda_edd (Lbol/Ledd): {:.4f} + {:.4f} -{:.4f}'.format(lambda_edd50,lambda_edd84-lambda_edd50,lambda_edd50-lambda_edd16))
    print('theta: {:.4f} + {:.4f} -{:.4f}'.format(theta50,theta84-theta50,theta50-theta16))

    

    import corner 
    samples = post_df[['log10Mbh',"log10Mdot","logL"]].values

    samples = np.concatenate((samples,lambda_edd.reshape(-1,1)), axis=1) 

    variable_LaTex = {**variable_LaTex,**{'logL':'logL', 'lambdaEdd':r'$\lambda_{Edd}$'}} 
    
    figure = corner.corner(samples,labels=list(variable_LaTex.values()),show_titles=True,quantiles=[0.16,0.50,0.84])
    figure.savefig(outdir+'corner.png')
    figure.show()
    

    bf_all = np.zeros((Nsample,xplot.shape[0]))

    for i in range(Nsample):
        if isinstance(Mbh,np.ndarray) and Mbh.shape[0] > 1:
            Mtmp = Mbh[i]
        else:
            Mtmp = Mbh
        if isinstance(Mdot,np.ndarray) and Mdot.shape[0] > 1:
            Mdottmp = Mdot[i]
        else:
            Mdottmp = Mdot
        if isinstance(theta,np.ndarray) and theta.shape[0] > 1:
            thetatmp = theta[i]
        else:
            thetatmp = theta
        bf = disc_luminosity(xplot,Mtmp,Mdottmp,thetatmp*np.pi/180)
        bf_all[i,:] = bf

    bf_median,bf_16,bf_84 = np.percentile(bf_all,[50,16,84],axis=0)

    fig = plt.figure()
    ax  = fig.add_subplot()
    ax.errorbar(np.log10(xdata),np.log10(ydata),yerr=yerr/(ydata*np.log(10)),fmt='.c',ms=10)
    ax.plot(np.log10(xplot),bf_median,'k',lw=3)
    #ax.plot(np.log10(xplot),bf_fix,'m',lw=3)
    ax.fill_between(np.log10(xplot),bf_84,y2=bf_median,color='k',alpha=0.25)
    ax.fill_between(np.log10(xplot),bf_median,y2=bf_16,color='k',alpha=0.25)
    
    ax.set_xlabel(r'$\nu~[Hz]$')
    ax.set_ylabel(r'$\nu L_{\nu}~[erg~Hz/s]$')

    #ax.set_ylim(44.5,47)

    data = np.concatenate((np.log10(xplot).reshape(-1,1),
                           bf_median.reshape(-1,1),
                           bf_84.reshape(-1,1),
                           bf_16.reshape(-1,1)),axis=1)

    np.savetxt(outdir+'best_fit_data.txt',data)

    fig.show()
    fig.savefig(outdir+'bestfit.png')


    return [Mbh50,Mdot50,lambda_edd50,theta50]


def calling_all(xdata,ydata,yerr,outdir,hyperpar_dict,variable_quantities,\
                fixed_quantities,variable_LaTex,hyperpar_disc_dict,which_sampler):
    """
    This function runs the parameter estimation and analysis for the accretion disc model.

    Parameters
    ----------
    xdata : np.ndarray
        Array of x-data. Frequency.
    ydata : np.ndarray
        Array of y-data. nu*Lnu
    yerr : np.ndarray
        Array of y-errors.
    outdir : str
        Output directory.
    hyperpar_dict : dict
        Hyperparameter dictionary for the sampler.
    variable_quantities : dict
        Dictionary of variable quantities.
    fixed_quantities : dict
        Dictionary of fixed quantities.
    variable_LaTex : dict
        Dictionary of variable quantities with LaTeX labels.
    hyperpar_disc_dict : dict
        Hyperparameter dictionary for the disc model.
    which_sampler : str
        Which sampler to use.

    Returns
    -------
    post_df : pandas.DataFrame
        Posterior samples as a pandas DataFrame.
    best : list
        Best-fit values for the parameters.

    """
    if which_sampler == 'CPnest':
        post_df = run_CPnest(xdata,ydata,yerr,outdir,hyperpar_dict,variable_quantities,\
                    fixed_quantities,hyperpar_disc_dict)
    elif which_sampler =='raynest':
        post_df = run_raynest(xdata,ydata,yerr,outdir,hyperpar_dict,variable_quantities,\
                    fixed_quantities,hyperpar_disc_dict)
    elif which_sampler =='nessai':
        post_df = run_nessai(xdata,ydata,yerr,outdir,hyperpar_dict,variable_quantities,\
                    fixed_quantities,hyperpar_disc_dict)
    else:
        raise ValueError('Unknown sampler {}'.format(which_sampler))


    best = analysis(post_df,xdata,ydata,yerr,outdir,fixed_quantities,variable_LaTex,hyperpar_disc_dict)
    return post_df,best



#if __name__=='__main__':




